from .abyss import *
from .achievement import *
from .artifact import *
from .book import *
from .changelog import *
from .character import *
from .character_fetter import *
from .domain import *
from .food import *
from .furniture import *
from .material import *
from .monster import *
from .name_card import *
from .quest import *
from .tcg import *
from .upgrade import *
from .weapon import *



from typing import Any

from pydantic import BaseModel, Field, field_validator

from ..utils import remove_html_tags

__all__ = (
    "WeaponAscensionMaterial",
    "WeaponPromoteCostItem",
    "WeaponPromote",
    "WeaponBaseStat",
    "WeaponUpgrade",
    "WeaponAffixUpgrade",
    "WeaponAffix",
    "WeaponDetail",
    "Weapon",
)


class WeaponAscensionMaterial(BaseModel):
    id: int
    rarity: int


class WeaponPromoteCostItem(BaseModel):
    id: int
    amount: int


class WeaponPromoteStat(BaseModel):
    id: str
    value: float


class WeaponPromote(BaseModel):
    unlock_max_level: int = Field(alias="unlockMaxLevel")
    promote_level: int = Field(alias="promoteLevel")
    cost_items: list[WeaponPromoteCostItem] | None = Field(None, alias="costItems")
    coin_cost: int | None = Field(None, alias="coinCost")
    required_player_level: int | None = Field(None, alias="requiredPlayerLevel")
    add_stats: list[WeaponPromoteStat] | None = Field(None, alias="addProps")

    @field_validator("cost_items", mode="before")
    def _convert_cost_items(cls, v: dict[str, int]) -> list[WeaponPromoteCostItem]:
        return [WeaponPromoteCostItem(id=int(k), amount=v[k]) for k in v]

    @field_validator("add_stats", mode="before")
    def _convert_add_stats(cls, v: dict[str, float]) -> list[WeaponPromoteStat]:
        return [WeaponPromoteStat(id=stat_id, value=v[stat_id]) for stat_id in v]


class WeaponBaseStat(BaseModel):
    prop_type: str | None = Field(None, alias="propType")
    init_value: float = Field(alias="initValue")
    growth_type: str = Field(alias="type")


class WeaponUpgrade(BaseModel):
    awaken_cost: list[int] = Field(alias="awakenCost")
    base_stats: list[WeaponBaseStat] = Field(alias="prop")
    promotes: list[WeaponPromote] = Field(alias="promote")


class WeaponAffixUpgrade(BaseModel):
    level: int
    description: str

    @field_validator("description", mode="before")
    def _format_description(cls, v: str) -> str:
        return remove_html_tags(v)


class WeaponAffix(BaseModel):
    name: str
    upgrades: list[WeaponAffixUpgrade] = Field(alias="upgrade")

    @field_validator("upgrades", mode="before")
    def _convert_upgrades(cls, v: dict[str, str]) -> list[WeaponAffixUpgrade]:
        return [WeaponAffixUpgrade(level=int(k), description=v[k]) for k in v]


class WeaponDetail(BaseModel):
    id: int
    rarity: int = Field(alias="rank")
    type: str
    name: str
    description: str
    icon: str
    story_id: int | None = Field(alias="storyId")
    affix: WeaponAffix | None
    route: str
    upgrade: WeaponUpgrade
    ascension_materials: list[WeaponAscensionMaterial] = Field(alias="ascension")

    @field_validator("story_id", mode="before")
    def _flatten_story_id(cls, v: list[int] | None) -> int | None:
        return v[0] if v else None

    @field_validator("icon", mode="before")
    def _convert_icon_url(cls, v: str) -> str:
        return f"https://api.ambr.top/assets/UI/{v}.png"

    @field_validator("affix", mode="before")
    def _convert_affix(cls, v: dict[str, dict[str, Any]] | None) -> WeaponAffix | None:
        if v is None:
            return None
        affix = list(v.values())[0]
        return WeaponAffix(**affix)

    @field_validator("ascension_materials", mode="before")
    def _convert_ascension_materials(cls, v: dict[str, int]) -> list[WeaponAscensionMaterial]:
        return [WeaponAscensionMaterial(id=int(k), rarity=v[k]) for k in v]


class Weapon(BaseModel):
    """
    Represents a weapon.

    Attributes
    ----------
    id: :class:`int`
        The weapon's ID.
    rarity: :class:`int`
        The weapon's rarity.
    type: :class:`str`
        The weapon's type.
    name: :class:`str`
        The weapon's name.
    icon: :class:`str`
        The weapon's icon.
    route: :class:`str`
        The weapon's route.
    """

    id: int
    rarity: int = Field(alias="rank")
    type: str
    name: str
    icon: str
    route: str

    @field_validator("icon", mode="before")
    def _convert_icon_url(cls, v: str) -> str:
        return f"https://api.ambr.top/assets/UI/{v}.png"



from pydantic import BaseModel, Field, field_validator

__all__ = (
    "BookVolume",
    "BookDetail",
    "Book",
)


class BookVolume(BaseModel):
    """
    Represents a book volume.

    Attributes
    ----------
    id: :class:`int`
        The book volume's ID.
    name: :class:`str`
        The book volume's name.
    description: :class:`str`
        The book volume's description.
    story_id: :class:`int`
        The book volume's story ID.
    """

    id: int
    name: str
    description: str
    story_id: int = Field(alias="storyId")


class BookDetail(BaseModel):
    """
    Represents a book detail.

    Attributes
    ----------
    id: :class:`int`
        The book's ID.
    name: :class:`str`
        The book's name.
    rarity: :class:`int`
        The book's rarity.
    icon: :class:`str`
        The book's icon.
    volumes: List[:class:`BookVolume`]
        The book's volumes.
    """

    id: int
    name: str
    rarity: int = Field(alias="rank")
    icon: str
    volumes: list[BookVolume] = Field(alias="volume")

    @field_validator("icon", mode="before")
    def _convert_icon_url(cls, v: str) -> str:
        return f"https://api.ambr.top/assets/UI/{v}.png"


class Book(BaseModel):
    """
    Represents a book.

    Attributes
    ----------
    id: :class:`int`
        The book's ID.
    name: :class:`str`
        The book's name.
    rarity: :class:`int`
        The book's rarity.
    icon: :class:`str`
        The book's icon.
    route: :class:`str`
        The book's route.
    """

    id: int
    name: str
    rarity: int = Field(alias="rank")
    icon: str
    route: str

    @field_validator("icon", mode="before")
    def _convert_icon_url(cls, v: str) -> str:
        return f"https://api.ambr.top/assets/UI/{v}.png"



from typing import Any

from pydantic import BaseModel, Field, field_validator

from ..utils import remove_html_tags, replace_pronouns

__all__ = ("Quest", "Task", "Quote", "Story", "CharacterFetter")


class Quest(BaseModel):
    id: int
    quest_title: str | None = Field(None, alias="questTitle")
    chapter_id: int = Field(alias="chapterId")
    chapter_title: str = Field(alias="chapterTitle")


class Task(BaseModel):
    type: str
    quest_list: list[Quest] = Field(alias="questList")


class Quote(BaseModel):
    """
    Represents a quote.

    Attributes
    ----------
    title: :class:`str`
        The quote's title.
    audio_id: :class:`str`
        The quote's audio ID.
    text: :class:`str`
        The quote's text.
    tips: Optional[:class:`str`]
        The quote's tips.
    tasks: List[:class:`Task`]
        The quote's tasks.
    """

    title: str
    audio_id: str = Field(alias="audio")
    text: str
    tips: str | None
    tasks: list[Task]

    @field_validator("text", mode="before")
    def _format_text(cls, v: str) -> str:
        return remove_html_tags(replace_pronouns(v))

    @field_validator("tips", mode="before")
    def _convert_empty_tips(cls, v: str) -> str | None:
        return v if v else None

    @field_validator("tasks", mode="before")
    def _convert_empty_tasks(cls, v: list[dict[str, Any]] | None) -> list[Task]:
        if v is None:
            return []
        return [Task(**task) for task in v]


class Story(BaseModel):
    title: str
    title2: str | None
    text: str
    text2: str | None
    tips: str | None

    @field_validator("text", mode="before")
    def _format_text(cls, v: str) -> str:
        return remove_html_tags(replace_pronouns(v))

    @field_validator("text2", mode="before")
    def _format_text2(cls, v: str | None) -> str | None:
        return remove_html_tags(replace_pronouns(v)) if v else None

    @field_validator("tips", mode="before")
    def _convert_empty_tips(cls, v: str) -> str | None:
        return v if v else None


class CharacterFetter(BaseModel):
    quotes: list[Quote]
    stories: list[Story] = Field(alias="story")

    @field_validator("quotes", mode="before")
    def _flatten_quotes(cls, v: dict[str, dict[str, Any]]) -> list[Quote]:
        return [Quote(**quote) for quote in v.values()]

    @field_validator("stories", mode="before")
    def _flatten_stories(cls, v: dict[str, dict[str, Any]]) -> list[Story]:
        return [Story(**story) for story in v.values()]



from typing import Any

from pydantic import BaseModel, Field, field_validator

from ..utils import remove_html_tags

__all__ = (
    "FoodSource",
    "FoodEffect",
    "FoodRecipe",
    "FoodDetail",
    "Food",
)


class FoodSource(BaseModel):
    name: str
    type: str


class FoodEffect(BaseModel):
    id: str
    description: str

    @field_validator("description", mode="before")
    def _format_description(cls, v: str) -> str:
        return remove_html_tags(v)


class FoodRecipe(BaseModel):
    effect_icon: str = Field(alias="effectIcon")
    effects: list[FoodEffect] = Field(alias="effect")

    @field_validator("effect_icon", mode="before")
    def _convert_effect_icon_url(cls, v: str) -> str:
        return f"https://api.ambr.top/assets/UI/{v}.png"

    @field_validator("effects", mode="before")
    def _convert_effects(cls, v: dict[str, str]) -> list[FoodEffect]:
        return [FoodEffect(id=item_id, description=v[item_id]) for item_id in v]


class FoodDetail(BaseModel):
    name: str
    description: str
    type: str
    recipe: FoodRecipe | bool
    map_mark: bool = Field(alias="mapMark")
    sources: list[FoodSource] = Field(alias="source")
    icon: str
    rarity: int = Field(alias="rank")
    route: str

    @field_validator("recipe", mode="before")
    def _convert_recipe(cls, v: bool | dict[str, Any]) -> FoodRecipe | bool:
        if isinstance(v, dict):
            return FoodRecipe(**v)
        return False

    @field_validator("icon", mode="before")
    def _convert_icon_url(cls, v: str) -> str:
        return f"https://api.ambr.top/assets/UI/{v}.png"


class Food(BaseModel):
    """
    Represents a food.

    Attributes
    ----------
    id: :class:`int`
        The food's ID.
    name: :class:`str`
        The food's name.
    type: :class:`str`
        The food's type.
    recipe: :class:`bool`
        Whether the food is a recipe.
    map_mark: :class:`bool`
        Whether the food has a map mark.
    icon: :class:`str`
        The food's icon.
    rarity: :class:`int`
        The food's rarity.
    route: :class:`str`
        The food's route.
    effect_icon: :class:`str`
        The food's effect icon.
    """

    id: int
    name: str
    type: str
    recipe: bool
    map_mark: bool = Field(alias="mapMark")
    icon: str
    rarity: int = Field(alias="rank")
    route: str
    effect_icon: str | None = Field(None, alias="effectIcon")

    @field_validator("recipe", mode="before")
    def _convert_recipe(cls, v: bool | None) -> bool:
        return bool(v)

    @field_validator("icon", mode="before")
    def _convert_icon_url(cls, v: str) -> str:
        return f"https://api.ambr.top/assets/UI/{v}.png"

    @field_validator("effect_icon", mode="before")
    def _convert_effect_icon_url(cls, v: str | None) -> str | None:
        return f"https://api.ambr.top/assets/UI/{v}.png" if v else None



import logging
from enum import Enum
from typing import Any, Final

from aiohttp_client_cache.backends.sqlite import SQLiteBackend
from aiohttp_client_cache.session import CachedSession

from .exceptions import AmbrAPIError, ConnectionTimeoutError, DataNotFoundError
from .models import (
    AbyssResponse,
    AchievementCategory,
    ArtifactSet,
    ArtifactSetDetail,
    Book,
    BookDetail,
    Changelog,
    Character,
    CharacterDetail,
    CharacterFetter,
    Domains,
    Food,
    FoodDetail,
    Furniture,
    FurnitureDetail,
    Material,
    MaterialDetail,
    Monster,
    MonsterDetail,
    Namecard,
    NamecardDetail,
    Quest,
    TCGCard,
    TCGCardDetail,
    UpgradeData,
    Weapon,
    WeaponDetail,
)
from .models.furniture import FurnitureSet, FurnitureSetDetail
from .utils import remove_html_tags

__all__ = ("AmbrAPI", "Language")


class Language(Enum):
    CHT = "cht"
    CHS = "chs"
    DE = "de"
    EN = "en"
    ES = "es"
    FR = "fr"
    ID = "id"
    JP = "jp"
    KR = "kr"
    PT = "pt"
    RU = "ru"
    TH = "th"
    VI = "vi"
    IT = "it"
    TR = "tr"


LOGGER_ = logging.getLogger("ambr.py")


class AmbrAPI:
    BASE_URL: Final[str] = "https://api.ambr.top/v2"

    def __init__(
        self,
        *,
        lang: Language = Language.EN,
        cache_ttl: int = 3600,
        headers: dict[str, Any] | None = None,
    ) -> None:
        self.lang = lang
        self.cache_ttl = cache_ttl

        self._session: CachedSession | None = None
        self._cache = SQLiteBackend("./.cache/ambr/aiohttp-cache.db", expire_after=cache_ttl)
        self._headers = headers or {"User-Agent": "ambr-py"}

    async def __aenter__(self) -> "AmbrAPI":
        await self.start()
        return self

    async def __aexit__(self, exc_type, exc, tb) -> None:  # noqa: ANN001
        await self.close()

    async def _request(
        self, endpoint: str, *, static: bool = False, use_cache: bool
    ) -> dict[str, Any]:
        """
        A helper function to make requests to the API.

        Parameters
        ----------
        endpoint: :class:`str`
            The endpoint to request from.
        static: :class:`bool`
            Whether to use the static endpoint or not. Defaults to ``False``.
        use_cache: :class:`bool`
            Whether to use the cache or not. Defaults to ``True``.

        Returns
        -------
        Dict[str, Any]
            The response from the API.
        """
        if self._session is None:
            msg = "Call `start` before making requests."
            raise RuntimeError(msg)

        if static:
            url = f"{self.BASE_URL}/static/{endpoint}"
        else:
            url = f"{self.BASE_URL}/{self.lang.value}/{endpoint}"

        LOGGER_.debug("Requesting %s...", url)

        if not use_cache:
            async with self._session.disabled(), self._session.get(url) as resp:
                if resp.status != 200:
                    self._handle_error(resp.status)
                data = await resp.json()
        else:
            async with self._session.get(url) as resp:
                if resp.status != 200:
                    self._handle_error(resp.status)
                data = await resp.json()

        return data

    def _handle_error(self, code: int) -> None:
        """
        A helper function to handle errors.
        """
        match code:
            case 404:
                raise DataNotFoundError
            case 522 | 524:
                raise ConnectionTimeoutError
            case _:
                raise AmbrAPIError(code)

    async def start(self) -> None:
        """
        Starts the client session.
        """
        self._session = CachedSession(headers=self._headers, cache=self._cache)

    async def close(self) -> None:
        """
        Closes the client session.
        """
        if self._session is not None:
            await self._session.close()

    async def fetch_achievement_categories(
        self, use_cache: bool = True
    ) -> list[AchievementCategory]:
        """
        Fetches all achievement categories.

        Returns
        -------
        List[:class:`AchievementCategory`]
            The achievement categories.
        """
        data = await self._request("achievement", use_cache=use_cache)
        return [
            AchievementCategory(**achievement_category)
            for achievement_category in data["data"].values()
        ]

    async def fetch_artifact_sets(self, use_cache: bool = True) -> list[ArtifactSet]:
        """
        Fetches all artifact sets.

        Returns
        -------
        List[:class:`ArtifactSet`]
            The artifact sets.
        """
        data = await self._request("reliquary", use_cache=use_cache)
        return [ArtifactSet(**artifact_set) for artifact_set in data["data"]["items"].values()]

    async def fetch_artifact_set_detail(self, id: int, use_cache: bool = True) -> ArtifactSetDetail:
        """
        Fetches an artifact set detail by ID.

        Parameters
        ----------
        id: :class:`int`
            The ID of the artifact set detail to fetch.

        Returns
        -------
        :class:`ArtifactSetDetail`
            The artifact set detail.
        """
        data = await self._request(f"reliquary/{id}", use_cache=use_cache)
        return ArtifactSetDetail(**data["data"])

    async def fetch_books(self, use_cache: bool = True) -> list[Book]:
        """
        Fetches all books.

        Returns
        -------
        List[:class:`Book`]
            The books.
        """
        data = await self._request("book", use_cache=use_cache)
        return [Book(**book) for book in data["data"]["items"].values()]

    async def fetch_book_detail(self, id: int, use_cache: bool = True) -> BookDetail:
        """
        Fetches a book detail by ID.

        Parameters
        ----------
        id: :class:`int`
            The ID of the book detail to fetch.

        Returns
        -------
        :class:`BookDetail`
            The book detail.
        """
        data = await self._request(f"book/{id}", use_cache=use_cache)
        return BookDetail(**data["data"])

    async def fetch_characters(self, use_cache: bool = True) -> list[Character]:
        """
        Fetches all characters.

        Returns
        -------
        List[:class:`Character`]
            The characters.
        """
        data = await self._request("avatar", use_cache=use_cache)
        return [Character(**character) for character in data["data"]["items"].values()]

    async def fetch_character_detail(self, id: str, use_cache: bool = True) -> CharacterDetail:
        """
        Fetches a character detail by ID.

        Parameters
        ----------
        id: :class:`str`
            The ID of the character detail to fetch.

        Returns
        -------
        :class:`CharacterDetail`
            The character detail.
        """
        data = await self._request(f"avatar/{id}", use_cache=use_cache)
        return CharacterDetail(**data["data"])

    async def fetch_character_fetter(self, id: str, use_cache: bool = True) -> CharacterFetter:
        """
        Fetches a character fetter by ID.

        Parameters
        ----------
        id: :class:`str`
            The ID of the character fetter to fetch.

        Returns
        -------
        :class:`CharacterFetter`
            The character fetter.
        """
        data = await self._request(f"avatarFetter/{id}", use_cache=use_cache)
        return CharacterFetter(**data["data"])

    async def fetch_foods(self, use_cache: bool = True) -> list[Food]:
        """
        Fetches all foods.

        Returns
        -------
        List[:class:`Food`]
            The foods.
        """
        data = await self._request("food", use_cache=use_cache)
        return [Food(**food) for food in data["data"]["items"].values()]

    async def fetch_food_detail(self, id: int, use_cache: bool = True) -> FoodDetail:
        """
        Fetches a food detail by ID.

        Parameters
        ----------
        id: :class:`int`
            The ID of the food detail to fetch.

        Returns
        -------
        :class:`FoodDetail`
            The food detail.
        """
        data = await self._request(f"food/{id}", use_cache=use_cache)
        return FoodDetail(**data["data"])

    async def fetch_furnitures(self, use_cache: bool = True) -> list[Furniture]:
        """
        Fetches all furnitures.

        Returns
        -------
        List[:class:`Furniture`]
            The furnitures.
        """
        data = await self._request("furniture", use_cache=use_cache)
        return [Furniture(**furniture) for furniture in data["data"]["items"].values()]

    async def fetch_furniture_detail(self, id: int, use_cache: bool = True) -> FurnitureDetail:
        """
        Fetches a furniture detail by ID.

        Parameters
        ----------
        id: :class:`int`
            The ID of the furniture detail to fetch.

        Returns
        -------
        :class:`FurnitureDetail`
            The furniture detail.
        """
        data = await self._request(f"furniture/{id}", use_cache=use_cache)
        return FurnitureDetail(**data["data"])

    async def fetch_furniture_sets(self, use_cache: bool = True) -> list[FurnitureSet]:
        """
        Fetches all furniture sets.

        Returns
        -------
        List[:class:`FurnitureSet`]
            The furniture sets.
        """
        data = await self._request("furnitureSuite", use_cache=use_cache)
        return [FurnitureSet(**furniture_set) for furniture_set in data["data"]["items"].values()]

    async def fetch_furniture_set_detail(
        self, id: int, use_cache: bool = True
    ) -> FurnitureSetDetail:
        """
        Fetches a furniture set detail by ID.

        Parameters
        ----------
        id: :class:`int`
            The ID of the furniture set detail to fetch.

        Returns
        -------
        :class:`FurnitureSetDetail`
            The furniture set detail.
        """
        data = await self._request(f"furnitureSuite/{id}", use_cache=use_cache)
        return FurnitureSetDetail(**data["data"])

    async def fetch_materials(self, use_cache: bool = True) -> list[Material]:
        """
        Fetches all materials.

        Returns
        -------
        List[:class:`Material`]
            The materials.
        """
        data = await self._request("material", use_cache=use_cache)
        return [Material(**material) for material in data["data"]["items"].values()]

    async def fetch_material_detail(self, id: int, use_cache: bool = True) -> MaterialDetail:
        """
        Fetches a material detail by ID.

        Parameters
        ----------
        id: :class:`int`
            The ID of the material detail to fetch.

        Returns
        -------
        :class:`MaterialDetail`
            The material detail.
        """
        data = await self._request(f"material/{id}", use_cache=use_cache)
        return MaterialDetail(**data["data"])

    async def fetch_monsters(self, use_cache: bool = True) -> list[Monster]:
        """
        Fetches all monsters.

        Returns
        -------
        List[:class:`Monster`]
            The monsters.
        """
        data = await self._request("monster", use_cache=use_cache)
        return [Monster(**monster) for monster in data["data"]["items"].values()]

    async def fetch_monster_detail(self, id: int, use_cache: bool = True) -> MonsterDetail:
        """
        Fetches a monster detail by ID.

        Parameters
        ----------
        id: :class:`int`
            The ID of the monster detail to fetch.

        Returns
        -------
        :class:`MonsterDetail`
            The monster detail.
        """
        data = await self._request(f"monster/{id}", use_cache=use_cache)
        return MonsterDetail(**data["data"])

    async def fetch_namecards(self, use_cache: bool = True) -> list[Namecard]:
        """
        Fetches all name cards.

        Returns
        -------
        List[:class:`NameCard`]
            The name cards.
        """
        data = await self._request("namecard", use_cache=use_cache)
        return [Namecard(**name_card) for name_card in data["data"]["items"].values()]

    async def fetch_namecard_detail(self, id: int, use_cache: bool = True) -> NamecardDetail:
        """
        Fetches a name card detail by ID.

        Parameters
        ----------
        id: :class:`int`
            The ID of the name card detail to fetch.

        Returns
        -------
        :class:`NameCardDetail`
            The name card detail.
        """
        data = await self._request(f"namecard/{id}", use_cache=use_cache)
        return NamecardDetail(**data["data"])

    async def fetch_quests(self, use_cache: bool = True) -> list[Quest]:
        """
        Fetches all quests.

        Returns
        -------
        List[:class:`Quest`]
            The quests.
        """
        data = await self._request("quest", use_cache=use_cache)
        return [Quest(**quest) for quest in data["data"]["items"].values()]

    async def fetch_tcg_cards(self, use_cache: bool = True) -> list[TCGCard]:
        """
        Fetches all TCG cards.

        Returns
        -------
        List[:class:`TCGCard`]
            The TCG cards.
        """
        data = await self._request("gcg", use_cache=use_cache)
        return [TCGCard(**tcg_card) for tcg_card in data["data"]["items"].values()]

    async def fetch_tcg_card_detail(self, id: int, use_cache: bool = True) -> TCGCardDetail:
        """
        Fetches a TCG card detail by ID.

        Parameters
        ----------
        id: :class:`int`
            The ID of the TCG card detail to fetch.

        Returns
        -------
        :class:`TCGCardDetail`
            The TCG card detail.
        """
        data = await self._request(f"gcg/{id}", use_cache=use_cache)
        return TCGCardDetail(**data["data"])

    async def fetch_weapons(self, use_cache: bool = True) -> list[Weapon]:
        """
        Fetches all weapons.

        Returns
        -------
        List[:class:`Weapon`]
            The weapons.
        """
        data = await self._request("weapon", use_cache=use_cache)
        return [Weapon(**weapon) for weapon in data["data"]["items"].values()]

    async def fetch_weapon_types(self, use_cache: bool = True) -> dict[str, str]:
        """
        Fetches all weapon types.

        Returns
        -------
        Dict[:class:`str`, :class:`str`]
            All of the weapon types.
        """
        data = await self._request("weapon", use_cache=use_cache)
        return data["data"]["types"]

    async def fetch_weapon_detail(self, id: int, use_cache: bool = True) -> WeaponDetail:
        """
        Fetches a weapon detail by ID.

        Parameters
        ----------
        id: :class:`int`
            The ID of the weapon detail to fetch.

        Returns
        -------
        :class:`WeaponDetail`
            The weapon detail.
        """
        data = await self._request(f"weapon/{id}", use_cache=use_cache)
        return WeaponDetail(**data["data"])

    async def fetch_domains(self, use_cache: bool = True) -> Domains:
        """
        Fetches all domains.

        Returns
        -------
        :class:`Domains`
            The domains.
        """
        data = await self._request("dailyDungeon", use_cache=use_cache)
        return Domains(**data["data"])

    async def fetch_changelogs(self, use_cache: bool = True) -> list[Changelog]:
        """
        Fetch changelogs from the API.

        Returns
        -------
        List[Changelog]
            A list of Changelog objects.
        """
        data = await self._request("changelog", static=True, use_cache=use_cache)
        changelogs: list[Changelog] = []
        for changelog_id, log in data["data"].items():
            changelogs.append(Changelog(id=int(changelog_id), **log))
        return changelogs

    async def fetch_upgrade_data(self, use_cache: bool = True) -> UpgradeData:
        """
        Fetch upgrade data from the API.

        Returns
        -------
        UpgradeData
            The upgrade data.
        """
        data = await self._request("upgrade", use_cache=use_cache)
        return UpgradeData(**data["data"])

    async def fetch_manual_weapon(self, use_cache: bool = True) -> dict[str, str]:
        """
        Fetch manual weapon data from the API.

        Returns
        -------
        Dict[str, str]
            The manual weapon data.
        """
        data = await self._request("manualWeapon", use_cache=use_cache)
        return data["data"]

    async def fetch_readable(self, id: str, use_cache: bool = True) -> str:
        """
        Fetch a readable from the API.

        Parameters
        ----------
        id: :class:`str`
            The ID of the readable to fetch.

        Returns
        -------
        :class:`str`
            The readable.
        """
        data = await self._request(f"readable/{id}", use_cache=use_cache)
        return remove_html_tags(data["data"])

    async def fetch_avatar_curve(
        self, use_cache: bool = True
    ) -> dict[str, dict[str, dict[str, float]]]:
        """
        Fetch avatar curve from the API.

        Returns
        -------
        Dict[str, Dict[str, Dict[str, float]]]
            The avatar curve.
        """
        data = await self._request("avatarCurve", static=True, use_cache=use_cache)
        return data["data"]

    async def fetch_weapon_curve(
        self, use_cache: bool = True
    ) -> dict[str, dict[str, dict[str, float]]]:
        """
        Fetch weapon curve from the API.

        Returns
        -------
        Dict[str, Dict[str, Dict[str, float]]]
            The weapon curve.
        """
        data = await self._request("weaponCurve", static=True, use_cache=use_cache)
        return data["data"]

    async def fetch_monster_curve(
        self, use_cache: bool = True
    ) -> dict[str, dict[str, dict[str, float]]]:
        """
        Fetch monster curve from the API.

        Returns
        -------
        Dict[str, Dict[str, Dict[str, float]]]
            The monster curve.
        """
        data = await self._request("monsterCurve", static=True, use_cache=use_cache)
        return data["data"]

    async def fetch_abyss_data(self, use_cache: bool = True) -> AbyssResponse:
        """
        Fetches abyss data from the API.

        Returns
        -------
        AbyssResponse
            The abyss data.
        """
        data = await self._request("tower", use_cache=use_cache)
        return AbyssResponse(**data["data"])



from typing import Any

from pydantic import BaseModel, Field, field_validator

from ..utils import remove_html_tags

__all__ = (
    "FurnitureRecipeInput",
    "FurnitureRecipe",
    "FurnitureDetail",
    "Furniture",
    "FurnitureSet",
    "FurnitureItem",
    "FurnitureSetDetail",
)


class FurnitureRecipeInput(BaseModel):
    id: int
    icon: str
    amount: int = Field(alias="count")

    @field_validator("icon", mode="before")
    def _convert_icon_url(cls, v: str) -> str:
        return f"https://api.ambr.top/assets/UI/{v}.png"


class FurnitureRecipe(BaseModel):
    exp: int
    time: int
    inputs: list[FurnitureRecipeInput] = Field(alias="input")

    @field_validator("inputs", mode="before")
    def _convert_inputs(cls, v: dict[str, dict[str, Any]]) -> list[FurnitureRecipeInput]:
        return [FurnitureRecipeInput(id=int(item_id), **v[item_id]) for item_id in v]


class FurnitureDetail(BaseModel):
    id: int
    name: str
    cost: int | None
    comfort: int | None
    rarity: int = Field(alias="rank")
    icon: str
    route: str
    categories: list[str]
    types: list[str]
    description: str
    recipe: FurnitureRecipe | None

    @field_validator("icon", mode="before")
    def _convert_icon_url(cls, v: str) -> str:
        return f"https://api.ambr.top/assets/UI/furniture/{v}.png"

    @field_validator("recipe", mode="before")
    def _convert_recipe(cls, v: dict[str, Any] | None) -> FurnitureRecipe | None:
        if v is None:
            return None
        return FurnitureRecipe(**v)


class Furniture(BaseModel):
    """
    Represents a furniture.

    Attributes
    ----------
    id: :class:`int`
        The furniture's ID.
    name: :class:`str`
        The furniture's name.
    cost: Optional[:class:`int`]
        The furniture's cost.
    comfort: Optional[:class:`int`]
        The furniture's comfort.
    rarity: :class:`int`
        The furniture's rarity.
    icon: :class:`str`
        The furniture's icon.
    route: :class:`str`
        The furniture's route.
    categories: List[:class:`str`]
        The furniture's categories.
    types: List[:class:`str`]
        The furniture's types.
    """

    id: int
    name: str
    cost: int | None
    comfort: int | None
    rarity: int = Field(alias="rank")
    icon: str
    route: str
    categories: list[str]
    types: list[str]

    @field_validator("icon", mode="before")
    def _convert_icon_url(cls, v: str) -> str:
        return f"https://api.ambr.top/assets/UI/furniture/{v}.png"


class FurnitureSet(BaseModel):
    id: int
    name: str
    icon: str
    route: str
    categories: list[str]
    types: list[str]

    @field_validator("icon", mode="before")
    def _convert_icon_url(cls, v: str) -> str:
        return f"https://api.ambr.top/assets/UI/furnitureSuite/{v}.png"

    @field_validator("categories", mode="before")
    def _convert_categories(cls, v: list[str] | None) -> list[str]:
        return v or []

    @field_validator("types", mode="before")
    def _convert_types(cls, v: list[str] | None) -> list[str]:
        return v or []


class FurnitureItem(BaseModel):
    id: int
    rarity: int = Field(alias="rank")
    icon: str

    @field_validator("icon", mode="before")
    def _convert_icon_url(cls, v: str) -> str:
        return f"https://api.ambr.top/assets/UI/furniture/{v}.png"


class FurnitureSetDetail(BaseModel):
    id: int
    name: str
    icon: str
    route: str
    categories: list[str]
    types: list[str]
    description: str
    furniture_items: list[FurnitureItem] = Field(alias="suiteItemList")

    @field_validator("icon", mode="before")
    def _convert_icon_url(cls, v: str) -> str:
        return f"https://api.ambr.top/assets/UI/furnitureSuite/{v}.png"

    @field_validator("categories", mode="before")
    def _convert_categories(cls, v: list[str] | None) -> list[str]:
        return v or []

    @field_validator("types", mode="before")
    def _convert_types(cls, v: list[str] | None) -> list[str]:
        return v or []

    @field_validator("description", mode="before")
    def _format_description(cls, v: str) -> str:
        return remove_html_tags(v)

    @field_validator("furniture_items", mode="before")
    def _convert_furniture_items(cls, v: dict[str, dict[str, Any]]) -> list[FurnitureItem]:
        return [FurnitureItem(id=int(item_id), **v[item_id]) for item_id in v]



import datetime
from typing import Any

from pydantic import BaseModel, Field, field_validator

from ..utils import remove_html_tags

__all__ = (
    "ChallengeTarget",
    "Chamber",
    "LeyLineDisorder",
    "Floor",
    "AbyssData",
    "Abyss",
    "AbyssEnemy",
    "AbyssResponse",
    "Blessing",
)


class Blessing(BaseModel):
    """
    Blessing model.

    Attributes:
        description (str): Description of the blessing.
        level_config_name (str): Level configuration name.
        visible (bool): Visibility status.
    """

    description: str
    level_config_name: str = Field(..., alias="levelConfigName")
    visible: bool

    @field_validator("description", mode="before")
    def _format_description(cls, v: str) -> str:
        return remove_html_tags(v)


class ChallengeTarget(BaseModel):
    """
    ChallengeTarget model.

    Attributes:
        type (str): Type of the challenge target.
        values (List[int]): List of values.
        formatted (str): Formatted challenge target.
    """

    type: str
    values: list[int]

    @property
    def formatted(self) -> str:
        return self.type.format("/".join(str(v) for v in self.values))


class Chamber(BaseModel):
    """
    Chamber model.

    Attributes:
        id (int): ID of the chamber.
        challenge_target (ChallengeTarget): Challenge target.
        enemy_level (int): Enemy level.
        wave_one_enemies (List[int]): List of enemies in the first wave.
        wave_two_enemies (Optional[List[int]]): List of enemies in the second wave.
    """

    id: int
    challenge_target: ChallengeTarget = Field(..., alias="challengeTarget")
    enemy_level: int = Field(..., alias="monsterLevel")
    wave_one_enemies: list[int] = Field(..., alias="firstMonsterList")
    wave_two_enemies: list[int] | None = Field(None, alias="secondMonsterList")


class LeyLineDisorder(BaseModel):
    """
    LeyLineDisorder model.

    Attributes:
        description (str): Description of the disorder.
        level_config_name (str): Level configuration name.
        visible (bool): Visibility status.
    """

    description: str
    level_config_name: str = Field(..., alias="levelConfigName")
    visible: bool

    @field_validator("description", mode="before")
    def _format_description(cls, v: str) -> str:
        return remove_html_tags(v)


class Floor(BaseModel):
    """
    Floor model.

    Attributes:
        id (int): ID of the floor.
        chambers (List[Chamber]): List of chambers.
        ley_line_disorders (List[LeyLineDisorder]): List of ley line disorders.
        override_enemy_level (int): Override enemy level.
        team_num (int): Number of teams.
    """

    id: int
    chambers: list[Chamber] = Field(..., alias="chamberList")
    ley_line_disorders: list[LeyLineDisorder] = Field(..., alias="leyLineDisorder")
    override_enemy_level: int = Field(..., alias="overrideMonsterLevel")
    team_num: int = Field(..., alias="teamNum")


class AbyssData(BaseModel):
    """
    AbyssData model.

    Attributes:
        open_time (Optional[datetime.datetime]): Opening time.
        floors (List[Floor]): List of floors.
    """

    open_time: datetime.datetime | None = Field(None, alias="openTime")
    floors: list[Floor] = Field(..., alias="floorList")

    @field_validator("open_time", mode="before")
    def _format_open_time(cls, v: int) -> datetime.datetime | None:
        return datetime.datetime.fromtimestamp(v) if v else None


class Abyss(BaseModel):
    """
    Abyss model.

    Attributes:
        id (int): ID of the abyss.
        open_time (datetime.datetime): Opening time.
        close_time (datetime.datetime): Closing time.
        blessing (Blessing): Blessing.
        abyss_corridor (AbyssData): Abyss corridor.
        abyssal_moon_spire (AbyssData): Abyssal moon spire.
    """

    id: int
    open_time: datetime.datetime = Field(..., alias="openTime")
    close_time: datetime.datetime = Field(..., alias="closeTime")
    blessing: Blessing
    abyss_corridor: AbyssData = Field(..., alias="entrance")
    abyssal_moon_spire: AbyssData = Field(..., alias="schedule")

    @field_validator("open_time", mode="before")
    def _format_open_time(cls, v: int) -> datetime.datetime:
        # example 1709258399
        return datetime.datetime.fromtimestamp(v)

    @field_validator("close_time", mode="before")
    def _format_close_time(cls, v: int) -> datetime.datetime:
        # example 1709258399
        return datetime.datetime.fromtimestamp(v)

    @field_validator("blessing", mode="before")
    def _format_blessing(cls, v: list[dict[str, Any]]) -> Blessing:
        return Blessing(**v[0])


class AbyssEnemyProperty(BaseModel):
    """
    AbyssEnemyProperty model.

    Attributes:
        initial_value (float): Initial value.
        type (str): Type of the property, e.g. "FIGHT_PROP_BASE_HP".
        growth_type (str): Growth type, e.g. "GROW_CURVE_HP".
    """

    initial_value: float = Field(..., alias="initValue")
    type: str = Field(..., alias="propType")
    growth_type: str = Field(..., alias="type")


class AbyssEnemy(BaseModel):
    """
    AbyssEnemy model.

    Attributes:
        icon (str): Icon URL.
        id (int): ID of the enemy.
        link (bool): Link status.
        name (str): Name of the enemy.
    """

    icon: str
    id: int
    link: bool
    name: str
    properties: list[AbyssEnemyProperty] = Field(..., alias="prop")

    @field_validator("icon", mode="before")
    def _convert_icon_url(cls, v: str) -> str:
        return f"https://api.ambr.top/assets/UI{'/monster' if 'MonsterIcon' in v else ''}/{v}.png"  # noqa: E501

    @field_validator("properties", mode="before")
    def _convert_properties(cls, v: list[dict[str, Any]]) -> list[AbyssEnemyProperty]:
        return [AbyssEnemyProperty(**prop) for prop in v]


class AbyssResponse(BaseModel):
    """
    AbyssResponse model.

    Attributes:
        enemies (Dict[str, AbyssEnemy]): Dictionary of abyss enemies.
        abyss_items (List[Abyss]): List of abyss items.
    """

    enemies: dict[str, AbyssEnemy] = Field(..., alias="monsterList")
    abyss_items: list[Abyss] = Field(..., alias="items")

    @field_validator("enemies", mode="before")
    def _convert_enemies(cls, v: dict[str, dict[str, Any]]) -> dict[str, AbyssEnemy]:
        return {item_id: AbyssEnemy(**v[item_id]) for item_id in v}

    @field_validator("abyss_items", mode="before")
    def _convert_abyss_items(cls, v: dict[str, dict[str, Any]]) -> list[Abyss]:
        result: list[Abyss] = []
        for item_id in v:
            v[item_id]["openTime"] = v[item_id]["schedule"]["openTime"]
            result.append(Abyss(**v[item_id]))
        return result



from pydantic import BaseModel, Field, field_validator

__all__ = ("Quest",)


class Quest(BaseModel):
    """
    Represents a quest.

    Attributes
    ----------
    id: :class:`int`
        The quest's ID.
    type: :class:`str`
        The quest's type.
    chapter_num: Optional[:class:`str`]
        The quest's chapter number.
    chapter_title: :class:`str`
        The quest's chapter title.
    chapter_icon: Optional[:class:`str`]
        The quest's chapter icon.
    chapter_image_title: :class:`str`
        The quest's chapter image title.
    route: :class:`str`
        The quest's route.
    chapter_count: :class:`int`
        The quest's chapter count.
    """

    id: int
    type: str | None
    chapter_num: str | None = Field(alias="chapterNum")
    chapter_title: str = Field(alias="chapterTitle")
    chapter_icon: str | None = Field(alias="chapterIcon")
    chapter_image_title: str | None = Field(alias="chapterImageTitle")
    route: str
    chapter_count: int = Field(alias="chapterCount")

    @field_validator("chapter_icon", mode="before")
    def _convert_icon_url(cls, v: str) -> str:
        return f"https://api.ambr.top/assets/UI/{v}.png"



from pydantic import BaseModel, Field, field_validator

__all__ = (
    "Item",
    "Changelog",
)


class Item(BaseModel):
    category: str
    ids: list[str]


class Changelog(BaseModel):
    """
    Represents a change log.

    Attributes
    ----------
    id: :class:`int`
        The change log's ID.
    version: :class:`str`
        The change log's version.
    items: List[:class:`Item`]
        The change log's items.
    beta: :class:`bool`
        Whether the change log is for beta.
    """

    id: int
    version: str
    items: list[Item]
    beta: bool = Field(False)

    @field_validator("items", mode="before")
    def _convert_items(cls, v: dict[str, list[str]]) -> list[Item]:
        return [Item(category=k, ids=v) for k, v in v.items()]  # type: ignore



from typing import Any

from pydantic import BaseModel, Field, field_validator

__all__ = (
    "AchievementReward",
    "AchievementDetail",
    "Achievement",
    "AchievementCategory",
)


class AchievementReward(BaseModel):
    """
    Represents an achievement reward.

    Attributes
    ----------
    rarity: :class:`int`
        The achievement reward's rarity.
    amount: :class:`int`
        The achievement reward's amount.
    icon: :class:`str`
        The achievement reward's icon.
    """

    rarity: int = Field(alias="rank")
    amount: int = Field(alias="count")
    icon: str

    @field_validator("icon", mode="before")
    def _convert_icon_url(cls, v: str) -> str:
        return f"https://api.ambr.top/assets/UI/{v}.png"


class AchievementDetail(BaseModel):
    """
    Represents an achievement detail.

    Attributes
    ----------
    id: :class:`int`
        The achievement's ID.
    title: :class:`str`
        The achievement's title.
    description: :class:`str`
        The achievement's description.
    rewards: List[:class:`AchievementReward`]
        The achievement's rewards.
    """

    id: int
    title: str
    description: str
    rewards: list[AchievementReward]

    @field_validator("rewards", mode="before")
    def _convert_rewards(cls, v: dict[str, dict[str, Any]]) -> list[AchievementReward]:
        return [AchievementReward(**v[item_id]) for item_id in v]


class Achievement(BaseModel):
    """
    Represents an achievement.

    Attributes
    ----------
    id: :class:`int`
        The achievement's ID.
    order: :class:`int`
        The achievement's order.
    details: List[:class:`AchievementDetail`]
        The achievement's details.
    """

    id: int
    order: int
    details: list[AchievementDetail]


class AchievementCategory(BaseModel):
    """
    Represents an achievement category.

    Attributes
    ----------
    id: :class:`int`
        The achievement category's ID.
    name: :class:`str`
        The achievement category's name.
    order: :class:`int`
        The achievement category's order.
    icon: :class:`str`
        The achievement category's icon.
    achievements: List[:class:`Achievement`]
        The achievement category's achievements.
    """

    id: int
    name: str
    order: int
    icon: str
    achievements: list[Achievement] = Field(alias="achievementList")

    @field_validator("icon", mode="before")
    def _convert_icon_url(cls, v: str) -> str:
        return f"https://api.ambr.top/assets/UI/{v}.png"

    @field_validator("achievements", mode="before")
    def _convert_achievements(cls, v: dict[str, dict[str, Any]]) -> list[Achievement]:
        return [Achievement(**v[achievement_id]) for achievement_id in v]



from typing import Any

from pydantic import BaseModel, Field, field_validator

from ..utils import remove_html_tags

__all__ = (
    "MonsterReward",
    "MonsterEntry",
    "MonsterDetail",
    "Monster",
)


class MonsterReward(BaseModel):
    id: int
    rarity: int = Field(alias="rank")
    icon: str
    count: float | None = Field(None)

    @field_validator("icon", mode="before")
    def _convert_icon_url(cls, v: str) -> str:
        return f"https://api.ambr.top/assets/UI/{v}.png"


class MonsterEntry(BaseModel):
    id: int
    type: str
    rewards: list[MonsterReward] = Field(alias="reward")

    @field_validator("rewards", mode="before")
    def _convert_rewards(cls, v: dict[str, dict[str, Any]] | None) -> list[MonsterReward]:
        return [MonsterReward(id=int(item_id), **v[item_id]) for item_id in v] if v else []


class MonsterDetail(BaseModel):
    id: int
    name: str
    type: str
    icon: str
    route: str
    title: str | None
    special_name: str | None = Field(alias="specialName")
    description: str
    entries: list[MonsterEntry]

    @field_validator("icon", mode="before")
    def _convert_icon_url(cls, v: str) -> str:
        return f"https://api.ambr.top/assets/UI{'/monster' if 'MonsterIcon' in v else ''}/{v}.png"  # noqa: E501

    @field_validator("entries", mode="before")
    def _convert_entries(cls, v: dict[str, dict[str, Any]]) -> list[MonsterEntry]:
        return [MonsterEntry(**v[item_id]) for item_id in v]

    @field_validator("description", mode="before")
    def _format_description(cls, v: str) -> str:
        return remove_html_tags(v)


class Monster(BaseModel):
    """
    Represents a living being.

    Attributes
    ----------
    id: :class:`int`
        The living being's ID.
    name: :class:`str`
        The living being's name.
    type: :class:`str`
        The living being's type.
    icon: :class:`str`
        The living being's icon.
    route: :class:`str`
        The living being's route.
    """

    id: int
    name: str
    type: str
    icon: str
    route: str

    @field_validator("icon", mode="before")
    def _convert_icon_url(cls, v: str) -> str:
        if "MonsterIcon" in v:
            return f"https://api.ambr.top/assets/UI/monster/{v}.png"
        else:
            return f"https://api.ambr.top/assets/UI/{v}.png"



from enum import IntEnum, StrEnum
from typing import Any

from pydantic import BaseModel, Field, field_validator

from ..utils import remove_html_tags

__all__ = (
    "Element",
    "Birthday",
    "ExtraLevelType",
    "TalentExtraLevel",
    "Constellation",
    "TalentType",
    "TalentUpgradeItem",
    "TalentUpgrade",
    "Talent",
    "AscensionMaterial",
    "CharacterPromoteStat",
    "CharacterPromoteMaterial",
    "CharacterPromote",
    "CharacterBaseStat",
    "CharacterUpgrade",
    "CharacterCV",
    "CharacterInfo",
    "CharacterDetail",
    "Character",
)


class Element(StrEnum):
    ANEMO = "Wind"
    GEO = "Rock"
    ELECTRO = "Electric"
    PYRO = "Fire"
    HYDRO = "Water"
    CRYO = "Ice"
    DENDRO = "Grass"


class Birthday(BaseModel):
    month: int
    day: int


class ExtraLevelType(IntEnum):
    NORMAL = 1
    ULTIMATE = 9
    SKILL = 2


class TalentExtraLevel(BaseModel):
    talent_type: ExtraLevelType = Field(alias="talentIndex")
    extra_level: int = Field(alias="extraLevel")


class Constellation(BaseModel):
    name: str
    description: str
    extra_level: TalentExtraLevel | None = Field(alias="extraData")
    icon: str

    @field_validator("description", mode="before")
    def _format_description(cls, v: str) -> str:
        return remove_html_tags(v)

    @field_validator("extra_level", mode="before")
    def _convert_extra_level(cls, v: dict[str, dict[str, Any]] | None) -> TalentExtraLevel | None:
        return TalentExtraLevel(**v["addTalentExtraLevel"]) if v else None

    @field_validator("icon", mode="before")
    def _convert_icon_url(cls, v: str) -> str:
        return f"https://api.ambr.top/assets/UI/{v}.png"


class TalentType(IntEnum):
    NORMAL = 0
    ULTIMATE = 1
    PASSIVE = 2


class TalentUpgradeItem(BaseModel):
    id: int
    amount: int


class TalentUpgrade(BaseModel):
    level: int
    cost_items: list[TalentUpgradeItem] | None = Field(None, alias="costItems")
    mora_cost: int | None = Field(None, alias="coinCost")
    description: list[str]
    params: list[int | float]

    @field_validator("cost_items", mode="before")
    def _convert_cost_items(cls, v: dict[str, int] | None) -> list[TalentUpgradeItem] | None:
        return [TalentUpgradeItem(id=int(k), amount=v[k]) for k in v] if v else None


class Talent(BaseModel):
    type: TalentType
    name: str
    description: str
    icon: str
    upgrades: list[TalentUpgrade] | None = Field(None, alias="promote")
    cooldown: float | None = Field(None)
    cost: int | None = Field(None)

    @field_validator("description", mode="before")
    def _format_description(cls, v: str) -> str:
        return remove_html_tags(v)

    @field_validator("icon", mode="before")
    def _convert_icon_url(cls, v: str) -> str:
        return f"https://api.ambr.top/assets/UI/{v}.png"

    @field_validator("upgrades", mode="before")
    def _convert_upgrades(cls, v: dict[str, dict[str, Any]]) -> list[TalentUpgrade]:
        return [TalentUpgrade(**upgrade) for upgrade in v.values()]


class AscensionMaterial(BaseModel):
    id: int
    rarity: int


class CharacterPromoteStat(BaseModel):
    id: str
    value: float


class CharacterPromoteMaterial(BaseModel):
    id: int
    count: int


class CharacterPromote(BaseModel):
    promote_level: int = Field(alias="promoteLevel")
    unlock_max_level: int = Field(alias="unlockMaxLevel")
    cost_items: list[CharacterPromoteMaterial] | None = Field(None, alias="costItems")
    add_stats: list[CharacterPromoteStat] | None = Field(None, alias="addProps")
    required_player_level: int | None = Field(None, alias="requiredPlayerLevel")
    coin_cost: int | None = Field(None, alias="coinCost")

    @field_validator("cost_items", mode="before")
    def _convert_cost_items(cls, v: dict[str, int]) -> list[CharacterPromoteMaterial]:
        return [CharacterPromoteMaterial(id=int(item_id), count=v[item_id]) for item_id in v]

    @field_validator("add_stats", mode="before")
    def _convert_add_stats(cls, v: dict[str, float]) -> list[CharacterPromoteStat]:
        return [CharacterPromoteStat(id=stat_id, value=v[stat_id]) for stat_id in v]


class CharacterBaseStat(BaseModel):
    prop_type: str = Field(alias="propType")
    init_value: float = Field(alias="initValue")
    growth_type: str = Field(alias="type")


class CharacterUpgrade(BaseModel):
    base_stats: list[CharacterBaseStat] = Field(alias="prop")
    promotes: list[CharacterPromote] = Field(alias="promote")


class CharacterCV(BaseModel):
    lang: str
    va: str


class CharacterInfo(BaseModel):
    title: str
    detail: str
    constellation: str
    native: str
    cv: list[CharacterCV]

    @field_validator("cv", mode="before")
    def _convert_cv(cls, v: dict[str, str]) -> list[CharacterCV]:
        return [CharacterCV(lang=lang, va=v[lang]) for lang in v]


class CharacterDetail(BaseModel):
    id: str
    rarity: int = Field(alias="rank")
    name: str
    element: Element
    weapon_type: str = Field(alias="weaponType")
    icon: str
    birthday: Birthday
    release: int | None = Field(None)
    route: str
    info: CharacterInfo = Field(alias="fetter")
    upgrade: CharacterUpgrade
    ascension_materials: list[AscensionMaterial] = Field(alias="ascension")
    talents: list[Talent] = Field(alias="talent")
    constellations: list[Constellation] = Field(alias="constellation")
    beta: bool = Field(False)
    special_stat: str = Field(alias="specialProp")
    region: str

    @field_validator("id", mode="before")
    def _stringify_id(cls, v: int) -> str:
        return str(v)

    @field_validator("icon", mode="before")
    def _convert_icon_url(cls, v: str) -> str:
        return f"https://api.ambr.top/assets/UI/{v}.png"

    @field_validator("birthday", mode="before")
    def _convert_birthday(cls, v: list[int]) -> Birthday:
        return Birthday(month=v[0], day=v[1])

    @field_validator("ascension_materials", mode="before")
    def _convert_ascension_materials(cls, v: dict[str, int]) -> list[AscensionMaterial]:
        return [AscensionMaterial(id=int(item_id), rarity=v[item_id]) for item_id in v]

    @field_validator("talents", mode="before")
    def _convert_talents(cls, v: dict[str, dict[str, Any]]) -> list[Talent]:
        return [Talent(**talent) for talent in v.values()]

    @field_validator("constellations", mode="before")
    def _convert_constellations(cls, v: dict[str, dict[str, Any]]) -> list[Constellation]:
        return [Constellation(**constellation) for constellation in v.values()]

    @property
    def gacha(self) -> str:
        """The character's gacha image."""
        return self.icon.replace("AvatarIcon", "Gacha_AvatarImg")


class Character(BaseModel):
    """
    Represents a character.

    Attributes
    ----------
    id: :class:`str`
        The character's ID.
    rarity: :class:`int`
        The character's rarity.
    name: :class:`str`
        The character's name.
    element: :class:`Element`
        The character's element.
    weapon_type: :class:`str`
        The character's weapon type.
    icon: :class:`str`
        The character's icon.
    birthday: List[:class:`str`]
        The character's birthday.
    release: :class:`int`
        The character's release date.
    route: :class:`str`
        The character's route.
    special_stat: :class:`str`
        The character's special stat, e.g. FIGHT_PROP_CRITICAL_HURT.
    region: :class:`str`
        The character's region.
    """

    id: str
    rarity: int = Field(alias="rank")
    name: str
    element: Element
    weapon_type: str = Field(alias="weaponType")
    icon: str
    birthday: Birthday
    release: int | None = Field(None)
    route: str
    beta: bool = Field(False)
    special_stat: str = Field(alias="specialProp")
    region: str

    @field_validator("id", mode="before")
    def _stringify_id(cls, v: int | str) -> str:
        return str(v)

    @field_validator("icon", mode="before")
    def _convert_icon_url(cls, v: str) -> str:
        return f"https://api.ambr.top/assets/UI/{v}.png"

    @field_validator("birthday", mode="before")
    def _convert_birthday(cls, v: list[int]) -> Birthday:
        return Birthday(month=v[0], day=v[1])

    @property
    def gacha(self) -> str:
        """The character's gacha image."""
        return self.icon.replace("AvatarIcon", "Gacha_AvatarImg")



from typing import Any

from pydantic import BaseModel, Field, field_validator

from ..utils import remove_html_tags, replace_placeholders

__all__ = (
    "CardTag",
    "DiceCost",
    "CardDictionary",
    "CardTalent",
    "TCGCardDetail",
    "TCGCard",
)


class CardTag(BaseModel):
    id: str
    name: str


class DiceCost(BaseModel):
    type: str
    amount: int = Field(alias="count")


class CardDictionary(BaseModel):
    id: str
    name: str
    params: dict[str, Any] | None = None
    description: str
    cost: list[DiceCost] = Field(None, alias="diceCost")

    @field_validator("name", mode="before")
    def _format_name(cls, v: str) -> str:
        return remove_html_tags(v)

    @field_validator("cost", mode="before")
    def _convert_cost(cls, v: dict[str, int] | None) -> list[DiceCost]:
        return [DiceCost(type=type_, count=count) for type_, count in v.items()] if v else []

    @field_validator("description", mode="before")
    def _format_description(cls, v: str, values: Any) -> str:
        params = values.data.get("params")
        if params:
            v = replace_placeholders(v, params)
        return remove_html_tags(v)


class CardTalent(BaseModel):
    id: str
    name: str
    params: dict[str, Any] | None
    description: str
    cost: list[DiceCost]
    tags: list[CardTag]
    icon: str
    sub_skills: dict[str, Any] | None = Field(None, alias="subSkills")

    @field_validator("description", mode="before")
    def _format_description(cls, v: str, values: Any) -> str:
        params = values.data.get("params")
        if params:
            v = replace_placeholders(v, params)
        return remove_html_tags(v)

    @field_validator("cost", mode="before")
    def _convert_cost(cls, v: dict[str, int] | None) -> list[DiceCost]:
        return [DiceCost(type=type_, count=count) for type_, count in v.items()] if v else []

    @field_validator("tags", mode="before")
    def _convert_tags(cls, v: dict[str, str] | None) -> list[CardTag]:
        return [CardTag(id=id_, name=name) for id_, name in v.items()] if v else []

    @field_validator("icon", mode="before")
    def _convert_icon_url(cls, v: str) -> str:
        return f"https://api.ambr.top/assets/UI/{v}.png"

    @property
    def small_icon(self) -> str:
        return self.icon.replace(".png", ".sm.png")


class TCGCardDetail(BaseModel):
    id: int
    name: str
    type: str
    tags: list[CardTag]
    props: dict[str, int] | None
    icon: str
    route: str
    story_title: str = Field(alias="storyTitle")
    story_detail: str = Field(alias="storyDetail")
    source: str
    dictionaries: list[CardDictionary] = Field(alias="dictionary")
    talents: list[CardTalent] = Field(alias="talent")

    @field_validator("tags", mode="before")
    def _convert_tags(cls, v: dict[str, str] | None) -> list[CardTag]:
        return [CardTag(id=id_, name=name) for id_, name in v.items()] if v else []

    @field_validator("icon", mode="before")
    def _convert_icon_url(cls, v: str) -> str:
        return f"https://api.ambr.top/assets/UI/gcg/{v}.png"

    @property
    def small_icon(self) -> str:
        return self.icon.replace(".png", ".sm.png")

    @field_validator("dictionaries", mode="before")
    def _convert_dictionaries(cls, v: dict[str, dict[str, Any]] | None) -> list[CardDictionary]:
        return [CardDictionary(id=item_id, **v[item_id]) for item_id in v] if v else []

    @field_validator("talents", mode="before")
    def _convert_talents(cls, v: dict[str, dict[str, Any]]) -> list[CardTalent]:
        return [CardTalent(id=item_id, **v[item_id]) for item_id in v]


class TCGCard(BaseModel):
    """
    Represents a TCG card.

    Attributes
    ----------
    id: :class:`int`
        The card's ID.
    name: :class:`str`
        The card's name.
    type: :class:`str`
        The card's type.
    tags: Dict[:class:`str`, :class:`str`]
        The card's tags.
    dice_cost: Dict[:class:`str`, :class:`int`]
        The card's properties.
    icon: :class:`str`
        The card's icon.
    route: :class:`str`
        The card's route.
    sort_order: :class:`int`
        The card's sort order.
    """

    id: int
    name: str
    type: str
    tags: list[CardTag]
    dice_cost: list[DiceCost] = Field(None, alias="props")
    icon: str
    route: str
    sort_order: int = Field(alias="sortOrder")

    @field_validator("tags", mode="before")
    def _convert_tags(cls, v: dict[str, str] | None) -> list[CardTag]:
        return [CardTag(id=id_, name=name) for id_, name in v.items()] if v else []

    @field_validator("dice_cost", mode="before")
    def _convert_dice_cost(cls, v: dict[str, int] | None) -> list[DiceCost]:
        return [DiceCost(type=type_, count=count) for type_, count in v.items()] if v else []

    @field_validator("icon", mode="before")
    def _convert_icon_url(cls, v: str) -> str:
        return f"https://api.ambr.top/assets/UI/gcg/{v}.png"

    @property
    def small_icon(self) -> str:
        return self.icon.replace(".png", ".sm.png")



import re
from typing import Any


def remove_html_tags(text: str) -> str:
    clean = re.compile(r"<.*?>|\{SPRITE_PRESET#[^\}]+\}")
    return re.sub(clean, "", text).replace("\\n", "\n")


def replace_placeholders(string: str, params: dict[str, Any]) -> str:
    for key, value in params.items():
        string = string.replace(f"$[{key}]", str(value))
    string = re.sub(r"\{SPRITE_PRESET#[^\}]+\}", "", string)
    return string


def replace_pronouns(text: str) -> str:
    female_pronoun_pattern = r"\{F#(.*?)\}"
    male_pronoun_pattern = r"\{M#(.*?)\}"

    female_pronoun_match = re.search(female_pronoun_pattern, text)
    male_pronoun_match = re.search(male_pronoun_pattern, text)

    if female_pronoun_match and male_pronoun_match:
        female_pronoun = female_pronoun_match.group(1)
        male_pronoun = male_pronoun_match.group(1)
        replacement = f"{female_pronoun}/{male_pronoun}"

        text = re.sub(female_pronoun_pattern, replacement, text)
        text = re.sub(male_pronoun_pattern, "", text)
        text = text.replace("#", "")

    return text



from typing import Any

from pydantic import BaseModel, Field, field_validator

from ..utils import remove_html_tags

__all__ = (
    "MaterialRecipe",
    "MaterialSource",
    "MaterialDetail",
    "Material",
)

WEEKDAYS = {
    "monday": 1,
    "tuesday": 2,
    "wednesday": 3,
    "thursday": 4,
    "friday": 5,
    "saturday": 6,
    "sunday": 0,
}


class MaterialRecipe(BaseModel):
    icon: str
    amount: int = Field(alias="count")

    @field_validator("icon", mode="before")
    def _convert_icon_url(cls, v: str) -> str:
        return f"https://api.ambr.top/assets/UI/{v}.png"


class MaterialSource(BaseModel):
    name: str
    type: str
    days: list[int] | None = Field(None)

    @field_validator("days", mode="before")
    def _convert_days(cls, v: list[str]) -> list[int]:
        return [WEEKDAYS[day] for day in v]


class MaterialDetail(BaseModel):
    name: str
    description: str
    type: str
    recipe: list[MaterialRecipe]
    map_mark: bool = Field(alias="mapMark")
    sources: list[MaterialSource] = Field(alias="source")
    icon: str
    rarity: int = Field(alias="rank")
    route: str

    @field_validator("description", mode="before")
    def _format_description(cls, v: str) -> str:
        return remove_html_tags(v)

    @field_validator("recipe", mode="before")
    def _convert_recipe(
        cls, v: bool | dict[str, dict[str, dict[str, Any]]]
    ) -> list[MaterialRecipe]:
        if isinstance(v, dict):
            recipe = list(v.values())[0]
            return [MaterialRecipe(**item) for item in recipe.values()]
        return []

    @field_validator("sources", mode="before")
    def _convert_sources(cls, v: list[dict] | None) -> list[MaterialSource]:
        return [MaterialSource(**item) for item in v] if v else []

    @field_validator("icon", mode="before")
    def _convert_icon_url(cls, v: str) -> str:
        return f"https://api.ambr.top/assets/UI/{v}.png"


class Material(BaseModel):
    """
    Represents a material.

    Attributes
    ----------
    id: :class:`int`
        The material's ID.
    name: :class:`str`
        The material's name.
    type: :class:`str`
        The material's type.
    recipe: :class:`bool`
        Whether the material is a recipe.
    map_mark: :class:`bool`
        Whether the material has a map mark.
    icon: :class:`str`
        The material's icon.
    rarity: :class:`int`
        The material's rarity.
    route: :class:`str`
        The material's route.
    """

    id: int
    name: str
    type: str
    recipe: bool
    map_mark: bool = Field(alias="mapMark")
    icon: str
    rarity: int = Field(alias="rank")
    route: str

    @field_validator("recipe", mode="before")
    def _convert_recipe(cls, v: bool | None) -> bool:
        return bool(v)

    @field_validator("icon", mode="before")
    def _convert_icon_url(cls, v: str) -> str:
        return f"https://api.ambr.top/assets/UI/{v}.png"



from typing import Any

from pydantic import BaseModel, Field, field_validator

__all__ = (
    "UpgradeItem",
    "Upgrade",
    "UpgradeData",
)


class UpgradeItem(BaseModel):
    id: int
    rarity: int


class Upgrade(BaseModel):
    id: str
    name: str | None = Field(None)
    icon: str
    items: list[UpgradeItem]

    @field_validator("icon", mode="before")
    def _convert_icon_url(cls, v: str) -> str:
        return f"https://api.ambr.top/assets/UI/{v}.png"

    @field_validator("items", mode="before")
    def _convert_items(cls, v: dict[str, int]) -> list[UpgradeItem]:
        return [UpgradeItem(id=int(k), rarity=v[k]) for k in v]


class UpgradeData(BaseModel):
    character: list[Upgrade] = Field(alias="avatar")
    weapon: list[Upgrade] = Field(alias="weapon")

    @field_validator("*", mode="before")
    def _convert_upgrade(cls, v: dict[str, dict[str, Any]]) -> list[Upgrade]:
        return [Upgrade(id=k, **v[k]) for k in v]



__all__ = ("AmbrAPIError", "DataNotFoundError", "ConnectionTimeoutError")


class AmbrAPIError(Exception):
    def __init__(self, code: int) -> None:
        self.code = code

    def __str__(self) -> str:
        return f"An error occurred while requesting the API, status code: {self.code}"


class DataNotFoundError(AmbrAPIError):
    def __init__(self) -> None:
        self.code = 404

    def __str__(self) -> str:
        return "Data not found"


class ConnectionTimeoutError(AmbrAPIError):
    def __init__(self) -> None:
        self.code = 522

    def __str__(self) -> str:
        return "Connection to the API timed out"



from pydantic import BaseModel, Field, field_validator

from ..utils import remove_html_tags

__all__ = (
    "NamecardDetail",
    "Namecard",
)


class NamecardDetail(BaseModel):
    id: int
    name: str
    rarity: int = Field(alias="rank")
    icon: str
    route: str
    description: str
    description_special: str = Field(alias="descriptionSpecial")
    source: str | None

    @field_validator("description", mode="before")
    def _format_description(cls, v: str) -> str:
        return remove_html_tags(v)

    @field_validator("icon", mode="before")
    def _convert_icon_url(cls, v: str) -> str:
        return f"https://api.ambr.top/assets/UI/namecard/{v}.png"

    @property
    def picture(self) -> str:
        return f"{self.icon.replace('NameCardIcon', 'NameCardPic')[:-4]}_P.png"


class Namecard(BaseModel):
    """
    Represents a namecard.

    Attributes
    ----------
    id: :class:`int`
        The name card's ID.
    name: :class:`str`
        The name card's name.
    type: :class:`str`
        The name card's type.
    rarity: :class:`int`
        The name card's rarity.
    icon: :class:`str`
        The name card's icon.
    route: :class:`str`
        The name card's route.
    """

    id: int
    name: str
    type: str
    rarity: int = Field(alias="rank")
    icon: str
    route: str

    @field_validator("icon", mode="before")
    def _convert_icon_url(cls, v: str) -> str:
        return f"https://api.ambr.top/assets/UI/namecard/{v}.png"

    @property
    def picture(self) -> str:
        return f"{self.icon.replace('NameCardIcon', 'NameCardPic')[:-4]}_P.png"



from typing import Any

from pydantic import BaseModel, Field, field_validator

__all__ = (
    "ArtifactAffix",
    "Artifact",
    "ArtifactSetDetail",
    "ArtifactSet",
)


class ArtifactAffix(BaseModel):
    """
    Represents an artifact set's set effect.

    Attributes
    ----------
    id: :class:`str`
        The effect's ID.
    effect: :class:`str`
        The effect's description.
    """

    id: str
    effect: str


class Artifact(BaseModel):
    """
    Represents an artifact.

    Attributes
    ----------
    pos: :class:`str`
        The position of the artifact.
    name: :class:`str`
        The name of the artifact.
    description: :class:`str`
        The description of the artifact.
    max_rarity: :class:`int`
        The maximum rarity of the artifact.
    icon: :class:`str`
        The name of the icon file for the artifact.
    """

    pos: str
    name: str
    description: str
    max_rarity: int = Field(alias="maxLevel")
    icon: str

    @field_validator("icon", mode="before")
    def _convert_icon_url(cls, v: str) -> str:
        return f"https://api.ambr.top/assets/UI/reliquary/{v}.png"


class ArtifactSetDetail(BaseModel):
    """
    Represents an artifact set detail.

    Attributes
    ----------
    id: :class:`int`
        The artifact set's ID.
    name: :class:`str`
        The artifact set's name.
    rarity_list: List[:class:`int`]
        The artifact set's rarity list.
    affix_list: List[:class:`ArtifactAffix`]
        The artifact set's affix list.
    icon: :class:`str`
        The artifact set's icon.
    route: :class:`str`
        The artifact set's route.
    artifacts: List[:class:`Artifact`]
        Artifacts that belong to the artifact set.
    """

    id: int
    name: str
    rarity_list: list[int] = Field(alias="levelList")
    affix_list: list[ArtifactAffix] = Field(alias="affixList")
    icon: str
    route: str
    artifacts: list[Artifact] = Field(alias="suit")

    @field_validator("affix_list", mode="before")
    def _convert_affix_list(cls, v: dict[str, str]) -> list[ArtifactAffix]:
        return [ArtifactAffix(id=k, effect=v[k]) for k in v]

    @field_validator("icon", mode="before")
    def _convert_icon_url(cls, v: str) -> str:
        return f"https://api.ambr.top/assets/UI/reliquary/{v}.png"

    @field_validator("artifacts", mode="before")
    def _convert_artifacts(cls, v: dict[str, dict[str, Any]]) -> list[Artifact]:
        return [Artifact(pos=artifact_pos, **v[artifact_pos]) for artifact_pos in v]


class ArtifactSet(BaseModel):
    """
    Represents an artifact set.

    Attributes
    ----------
    id: :class:`int`
        The artifact set's ID.
    name: :class:`str`
        The artifact set's name.
    rarity_list: List[:class:`int`]
        Obtainable rarities of the artifact set.
    affix_list: List[:class:`str`]
        The artifact set's set effect list.
    icon: :class:`str`
        The artifact set's icon.
    route: :class:`str`
        The artifact set's route.
    sort_order: :class:`int`
        The artifact set's sort order.
    """

    id: int
    name: str
    rarity_list: list[int] = Field(alias="levelList")
    affix_list: list[ArtifactAffix] = Field(alias="affixList")
    icon: str
    route: str
    sort_order: int = Field(alias="sortOrder")

    @field_validator("icon", mode="before")
    def _convert_icon_url(cls, v: str) -> str:
        return f"https://api.ambr.top/assets/UI/reliquary/{v}.png"

    @field_validator("affix_list", mode="before")
    def _convert_affix_list(cls, v: dict[str, str]) -> list[ArtifactAffix]:
        return [ArtifactAffix(id=k, effect=v[k]) for k in v]



from enum import IntEnum
from typing import Any

from pydantic import BaseModel, Field, field_validator

__all__ = (
    "City",
    "Domain",
    "Domains",
)


class City(IntEnum):
    MONDSTADT = 1
    LIYUE = 2
    INAZUMA = 3
    SUMERU = 4
    FONTAINE = 5


class DomainReward(BaseModel):
    id: int

    @property
    def icon(self) -> str:
        return f"https://api.ambr.top/assets/UI/UI_ItemIcon_{self.id}.png"


class Domain(BaseModel):
    id: int
    name: str
    rewards: list[DomainReward] = Field(alias="reward")
    city: City

    @field_validator("rewards", mode="before")
    def convert_rewards(cls, v: list[int]) -> list[DomainReward]:
        return [DomainReward(id=id_) for id_ in v]


class Domains(BaseModel):
    monday: list[Domain]
    tuesday: list[Domain]
    wednesday: list[Domain]
    thursday: list[Domain]
    friday: list[Domain]
    saturday: list[Domain]
    sunday: list[Domain]

    @staticmethod
    def _convert_domains(domains: dict[str, dict[str, Any]]) -> list[Domain]:
        return [Domain(**domain) for domain in domains.values()]

    @field_validator("*", mode="before")
    def convert_domains(cls, v: dict[str, dict[str, Any]]) -> list[Domain]:
        return cls._convert_domains(v)
